"""
This module contains functionality for i) solving (fitting, calibrating...) the soft 
maximin problem and ii) predicting from this solution (fitted model).

"""
import numpy as np
from ._smme import pga
from pysmme.transforms import iwt, RH

class _smme_dict(dict):
   def __init__(self, d):
      self._dict = d      
   def __repr__(self):
      out = self._dict["spec"] + ":" + "\n"
      j = 0
      for z in self._dict["zeta"]: # list of length num of z of np arrays of len endmod(z)
        out = out + "\n" + "{:<15} {:<8} {:<15} {:<8}".format("zeta = " + str(z), "DF", "Lambda", "Iter")
        for i in range(len(self._dict["df"][j])):
          df = int(self._dict["df"][j][i])
          lamb = self._dict["lamb"][j][i]
          iter = int(self._dict["iter"][j][i])
          out = out + "\n" + "{:<15} {:<8} {:<15} {:<8}".format("  ", df, f"{lamb:.5}", iter)
        j = j + 1
      return out
      
def softmaximin(
y,#, #G list or d array
x, #G list or d list or string for wavelets
zeta,
penalty,
alg,
nlamb = 30,
lamb_min_ratio = 1e-04,
lamb = None,
penalty_factor = None,
scale_y = 1,
reltol = 1e-05,
maxiter = 1000,
steps = 1,
btmax = 100,
c = 0.0001,
tau = 2,
M = 4,
nu = 1,
Lmin = 0,
lse = True,
nthreads = 4):
  
  r"""Function for solving the soft maximin estimation problem

  Parameters
  ----------
  y : list of arrays or array
    For a model with varying design across groups a list containing the :math:`G` group specific 
    response vectors of sizes :math:`n_i \times 1` . For a model with identical design 
    across :math:`G` groups, 
    an array of size :math:`n_1 \times\cdots\times n_d \times G` (:math:`d \in \{ 1, 2, 3\}`).
  x : list of arrays or string
    For a model with varying design across groups a list containing the 
    :math:`G` group specific design matrices of sizes :math:`n_i \times p`.  
    For a model with identical design across :math:`G` groups, either i) a list containing the 
    :math:`d \in \{ 1, 2, 3\}` marginal design matrices (tensor components) or ii) 
    a string indicating the type of wavelets to be used, see ``pysmme.transforms.wt`` for options. 
  zeta : array of strictly positive floats 
    Controls  the soft maximin approximation accuracy. When ``len(zeta) > 1`` 
    the procedure will distribute
    the computations using the   ``nthreads`` parameter below when openMP is available.
  penalty : string 
    Specifies the penalty type. Possible values are ``lasso, scad``.
  alg : string 
    Specifies the optimization algorithm. Possible values are ``npg, fista``.
  nlambda : strictly positive int 
    The number of ``lamb`` values to use when ``lamb`` is not specified.
  lamb_min_ratio : strictly positive float 
    Controls the minimum ``lamb`` value by setting the ratio bewtween 
    :math:`\lambda_{max}` -- the (data dependent) smallest value for which all 
    coefficients are zero --  and the smallest value of ``lamb``.
    Used when ``lamb`` is not specified.
  lamb : array of strictly positive floats 
    Penalty parameters.
  penalty_factor : np.array  
    Positive floats  that are multiplied with the parameters to allow for
    differential penalization on the these. Same size and shape as the model 
    coefficient container (array or vector).
  scale_y : strictly positive float 
    Scaling factor for the response   ``y``. To temper potential overflows.
  reltol : strictly positive float 
    Convergence tolerance for the proximal algorithm.
  maxiter : positive int
    The maximum number of  iterations
    allowed for each   ``lamb`` value, when  summing over all outer iterations
    for said   ``lamb``.
  steps : strictly positive int 
    The number of steps used in the multi-step adaptive lasso algorithm for 
    non-convex penalties. Automatically  set to 1 when   ``penalty = "lasso"``.
  btmax : strictly positive integer 
   The maximum number of backtracking steps allowed in each iteration. 
  c : strictly positive float 
    Used in the NPG algorithm. 
  tau : strictly positive float 
    Used to control the stepsize for NPG. 
  M : positive int
     The look back for the NPG. 
  nu : strictly positive float
    Ccontrols the stepsize in the proximal algorithm. A  value less than 1 will decrease 
    the stepsize and a value larger than one will increase it.
  Lmin : positive float 
    Controls the stepsize in the NPG algorithm. For the default  
    ``Lmin = 0`` the maximum step size is the same
    as for the FISTA algorithm.
  lse : bool 
    Indicates if log sum exp-loss is used.  TRUE is
    default and yields the loss below.
  nthreads : pos int
    The number of threads to use when  openMP  is available. 

  Returns
  -------  
  spec : string 
    Specifications of the model fitted by the function call.
  coef : list or np.array
   A :math:`p \times` ``nlamb`` matrix containing the
   estimates of the model coefficients for each   ``lamb``-value
   for which the procedure converged. When   ``len(zeta) > 1``
   a   ``len(zeta)``-list of such matrices.
  lamb : list or np.array
   The sequence of penalty values used
   in the estimation procedure for which the procedure converged.
   When   ``len(zeta) > 1`` a   ``len(zeta)``-list of such vectors.
  Obj : list or np.array
   The objective values for each
   iteration and each model for which the procedure converged.
   When   ``len(zeta) > 1`` a   ``len(zeta)``-list of such matrices.
  df : list or np.array 
   Vector containing the nonzero model coefficients (degrees of freedom) for each
   value of   ``lamb`` for which the procedure converged. When
   ``len(zeta) > 1`` a   ``len(zeta)``-list of such vectors.
  dimcoef : int or np.array
   Indicating the number :math:`p` of model parameters.
   For array data a vector giving the dimension of the model coefficient array.
  dimobs : int or np.array
    The number of observations. For array data a vector giving the number of  
    observations in each dimension.
  dimmodel : int or None
   The dimension of the array model. ``None`` for general models.
  diagnostics : dict 
   Key ``iter`` is a vector containing the number of  iterations for each
   ``lamb`` value for which the algorithm converged. When ``len(zeta) > 1`` a   
   ``len(zeta)``-list of such vectors. Key ``bt_iter``  is a  ``len(zeta)`` vector
   with total number of backtracking steps performed across all (converged) ``lamb`` values 
   for given ``zeta`` value. Key ``bt_enter`` is a  ``len(zeta)`` vector
   with total number of times backtracking is initiated across all (converged) ``lamb`` values 
   for given ``zeta`` value.

  Notes
  -----
  Consider modeling heterogeneous data :math:`\{y_1,\ldots, y_n\}` by dividing
  it into :math:`G` groups :math:`\mathbf{y}_g = (y_1, \ldots, y_{n_g})` ,
  :math:`g \in \{ 1,\ldots, G\}` and then using a linear model
  
  .. math:: \mathbf{y}_g = \mathbf{X}_gb_g + \epsilon_g, \  g \in \{1,\ldots, G\},
  
  to model the group response. Then :math:`b_g` is a group specific :math:`p\times 1`
  coefficient vector, :math:`\mathbf{X}_g` an :math:`n_g\times p` group design matrix and
  :math:`\epsilon_g` an :math:`n_g\times 1` error term. The objective is to estimate
  a common coefficient :math:`\beta` such that :math:`\mathbf{X}_g\beta` is a robust
  and good approximation to :math:`\mathbf{X}_gb_g` across groups.
 
  Following [1]_, this objective may be accomplished by
  solving the soft maximin estimation problem
  
  .. math:: \min_{\beta}\frac{1}{\zeta}\log\bigg(\sum_{g = 1}^G \exp(-\zeta \hat V_g(\beta))\bigg) + \lambda  \Vert\beta\Vert_1, \quad \zeta > 0,\lambda \geq 0.
  
  Here :math:`\zeta` essentially controls the amount of pooling across groups
  (:math:`\zeta \sim 0` effectively ignores grouping and pools observations) and
   
  .. math:: \hat V_g(\beta):=\frac{1}{n_g}(2\beta^\top \mathbf{X}_g^\top \mathbf{y}_g -\beta^\top \mathbf{X}_g^\top \mathbf{X}_g\beta),
  
  is the empirical explained variance, see [1]_ for more
  details and references.
 
  The function  ``softmaximin`` solves the soft maximin estimation problem in
  large scale settings for a sequence of penalty parameters
  :math:`\lambda_{max}>\ldots >\lambda_{min}>0` and a sequence of strictly positive
  softmaximin  parameters :math:`\zeta_1, \zeta_2,\ldots`.
 
  The implementation also solves the
  problem above with the penalty given by the SCAD penalty, using the multiple
  step adaptive lasso procedure to loop over the inner proximal algorithm.
 
  Two optimization algorithms  are implemented in the SMME packages;
  a non-monotone proximal gradient (NPG) algorithm and a fast iterative soft
  thresholding algorithm (FISTA).
 
  The implementation is particularly efficient for models where the design is
  identical across groups i.e. :math:`\mathbf{X}_g = \mathbf{X}`
  :math:`\forall g \in \{1, \ldots, G\}` in the following two cases:

  i) first if :math:`\mathbf{X}` has Kronecker (tensor) structure i.e. for marginal :math:`n_i\times p_i` design matrices :math:`\mathbf{M}_1,\ldots, \mathbf{M}_d`
  , :math:`d \in \{ 1, 2, 3\}`,

  .. math:: \mathbf{X} = \bigotimes_{i=1}^d \mathbf{M}_i 
  
  then ``y`` is a :math:`d + 1` dimensional response array
  and    ``x`` is a list containing the :math:`d` marginal matrices
  :math:`\mathbf{M}_1,\ldots, \mathbf{M}_d`. In this case  softmaximin solves
  the soft maximin problem using minimal memory by way of tensor optimized
  arithmetic, see also   ``RH``.

  ii) second, if the design matrix :math:`\mathbf{X}` is the inverse matrix of an
  orthogonal wavelet transform  then ``softmaximin``  will solve the soft maximin 
  problem given  ``x = str`` -- where ``str`` is a shorthand for the wavelet basis (see....) -- and 
  the :math:`d + 1` dimensional response array  ``y``. In this case  the  pyramid algorithm is used 
  to compute multiplications involving :math:`\mathbf{X}`.
 
  Note that when multiple values for :math:`\zeta` is provided it is  possible to
  distribute the computations across CPUs if openMP is available.
 
  References
  ----------
  .. [1] Lund, A., S. W. Mogensen and N. R. Hansen (2022). Soft Maximin Estimation for
      Heterogeneous Data. Scandinavian Journal of Statistics. url = https://doi.org/10.1111/sjos.12580
 
  Examples
  --------
  #Non-array data ##size of example

  >>> G = 3; 
  >>> n = np.array([65, 26, 13])
  >>> p = np.array([13, 5, 4])
  
  ##marginal design matrices (Kronecker components)

  >>> x = [None] * 3 
  >>> for i in range(len(x)):
  >>> x[i] = np.random.normal(0, 1, (n[i], p[i]))
       
  ##common features and effects

  >>> common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
  >>> common_effects = np.random.normal(size = np.prod(p)) * common_features
  
  ##group response

  >>> y = np.zeros((n[0], n[1], n[2], G))
  >>> for g in range(G):
  >>>     bg = np.random.normal(0, 0.1, np.prod(p)) * (1 - common_features) + common_effects
  >>>     mu = RH(x[2], RH(x[1], RH(x[0], np.reshape(bg, (p[0], p[1], p[2]), "F") )))
  >>>     y[:, :, :, g] = np.random.normal(0, 1, (n)) + mu
  
  ##fit model for range of lambda and zeta
  
  >>> zeta = np.array([0.1, 1, 10, 100])
  >>> fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
  >>> modelno = 10
  >>> zetano = 2
  >>> betahat = fit["coef"][zetano][:, modelno]
   
  >>> f, ax = plt.subplots(1)
  >>> ax.plot(common_effects, "r+")
  >>> ax.plot(betahat)
  >>> plt.show() 
   
  #Array data and wavelets
  ##size of example

  >>> set.seed(42)
  >>> G = 5; 
  >>> p = n = np.array([2**2, 2**3, 2**4])
  
  ##common features and effects
  
  >>> common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
  >>> common_effects = np.random.normal(size = np.prod(p)) * common_features
  
  ##group response

  >>> y = np.zeros((n[0], n[1], n[2], G))
  >>> for g in range(G):
  >>>     bg = np.random.normal(0, 0.1, np.prod(p)) * (1 - common_features) + common_effects
  >>>     mu = iwt(np.reshape(bg, (p[0], p[1], p[2]), "F"))
  >>>     y[:, :, :, g] = np.random.normal(0, 1, (n)) + mu
  
  ##fit model for range of lambda and zeta
  
  >>> zeta = np.array([0.1, 1, 10, 100])
  >>> fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
  >>> modelno = 10
  >>> zetano = 2
  >>> betahat = fit["coef"][zetano][:, modelno]
   
  >>> f, ax = plt.subplots(1)
  >>> ax.plot(common_effects, "r+")
  >>> ax.plot(betahat)
  >>> plt.show() 
  
  ##Non-array data
  ##size of example

  >>> G = 10
  >>> n = np.random.choice(np.arange(100,500,1), G) #sample(100:500, G); 
  >>> p = 60
  >>> x = [None] * G
  
  ##group design matrices

  >>> for i in range(len(x)):
  >>> x[i] = np.random.normal(0, 1, (n[i], p))
  
  ##common features and effects

  >>> common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
  >>> common_effects = np.random.normal(size = np.prod(p)) * common_features
  
  ##group response

  >>> y = [None] * G
  >>> for g in range(G):
  >>>     bg = np.random.normal(0, 0.5, np.prod(p)) * (1 - common_features) + common_effects
  >>>     mu = np.matmul(x[g], bg)
  >>>     y[g] = np.random.normal(0, 1, n[g]) + mu
  
  ##fit model for range of lamb and zeta
  
  >>> fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
  >>> betahat = fit["coef"]
  
  ##estimated common effects for specific lamb and zeta

  >>> modelno = 6 
  >>> zetano = 2
  >>> f, ax = plt.subplots(1)
  >>> ax.plot(common_effects, "r+")
  >>> ax.plot(betahat[zetano][:, modelno])
  >>> plt.show() 

  """

  try:
   num_of_zeta = len(zeta)
  except TypeError:
   num_of_zeta = 1
    
  Z = [] 
  if(type(y) == np.ndarray): #array data fixed design
    y = y * scale_y
    fix_array = True 
    G = y.shape[len(y.shape) - 1]
    dimglam = len(y.shape) - 1 #
    if(dimglam not in  [1, 2, 3]):
      return "Error: The dimension of an array model must be d = 1, 2 or 3!"
    if(type(x) ==  str): #wavelet design
      if(x not in ["haar", "d4", "??","mb4","fk4","d6","fk6", "d8","fk8", "la8",
                   "mb8","bl14","fk14", "d16","la16","mb16", "la20","bl20","fk22", "mb24"]): ##todo!!!!
        return "Error: The wavelet design is not correctly specified"
      if(np.mean(np.round(np.log2(y.shape[0:dimglam])) == np.log2(y.shape[0:dimglam])) != 1):
        return "Error: data is not dyadic so cannot use wavelet design"
      wave = True
      wf = x
      p1 = n1 = y.shape[0]
      if(dimglam == 1):
        p2 = n2 = 1
        p3 = n3 = 1
        J = int(np.log2(p1))
      elif(dimglam == 2):
        p2 = n2 = y.shape[1]
        p3 = n3 = 1
        J = int(np.log2(min(p1, p2)))
      else:
        p2 = n2 = y.shape[1]
        p3 = n3 = y.shape[2]
        J = int(np.log2(min(p1, p2, p3)))
      p = n = n1 * n2 * n3
      #x = [np.ones((2, 2)) * n1, np.ones((2, 2)) * n2, np.ones((2, 2)) * n3]
      x = [n1, n2, n3]
    elif(type(x) == list): ##custom fixed array tensor design
      if(len(x) != dimglam): 
        return "Error: For array data y the number of marginal design matrices in x must equal the dimension of the group data (d = 1, 2, or 3)"
      J = 0
      wave = False
      wf = "not used"
      if (dimglam == 1):
        x.append(np.ones((1, 1)))
        x.append(np.ones((1, 1)))
      if (dimglam == 2):
        x.append(np.ones((1, 1))) 
      n1 = x[0].shape[0]
      n2 = x[1].shape[0]
      n3 = x[2].shape[0]
      p1 = x[0].shape[1]
      p2 = x[1].shape[1]
      p3 = x[2].shape[1]
      n = n1 * n2 * n3
      p = p1 * p2 * p3
    else:
      return "Error: x not correctly specified"
   #structure array as G-list of n1xn2n3 matrices
    for i in range(G): 
      if(dimglam == 1):
        Z.append(np.reshape(y[:, i], (n1, n2 * n3) , order  = "F"))
      elif(dimglam == 2):
        Z.append(np.reshape(y[:,:, i], (n1, n2 * n3) , order  = "F"))
      elif(dimglam == 3):
        Z.append(np.reshape(y[:,:,:, i], (n1, n2 * n3) , order  = "F")) 
    del y  #is Z a copy?   
    if(penalty_factor == None):
      penalty_factor = np.ones((p1, p2 * p3), order  = "F")
    else:
      penalty_factor = np.reshape(penalty_factor, (p1, p2 * p3), order  = "F") 
  elif(type(y) == list): #varying design
    G = len(y)
    for i in range(G):
      Z.append(y[i] * scale_y)
    del y    
    fix_array = False
    wave = False
    dimglam = 0
    J = 0
    wf = "not used"
    if(len(x) != G):
      return "Error: The number of group design matrices in x must equal the number of the groups G in the data"
    #check to make sure y is compaitble with x in every gropu...todo
    n = 0
    for i in range(G):
      n = n + Z[i].shape[0]
    p = x[0].shape[1]
    if(penalty_factor == None):
      penalty_factor = np.ones(p) #np
  else:
    return "Error: y must be list or array"
  if(alg not in ["npg", "fista"]):
    return "Error: Algorithm must be correctly specified"
  if(c <= 0):
    return "Error: c must be strictly positive"
  if(Lmin < 0):
    return "Error: Lmin must be positive"
  #if(np.mean(zeta <= 0) > 0): todo!!
  #  return "all zetas must be strictly positive"
  if(penalty not in ["lasso", "scad"]):
    return "Error: Penalty must be correctly specified"
  if(np.min(penalty_factor) < 0):
    return "Error: penalty.factor must be positive"
  if(penalty_factor.size != p):
    return "Error: Number of elements in penalty.factor " + str(penalty_factor.size) + " is not equal to the number of coefficients " +  str(p)
  if(penalty == "lasso"):
    steps = 1
  if(lamb == None):
    makelamb = True
    lamb = np.zeros(nlamb) #fill with nas??
  else:
    lamb = scale_y * lamb
    makelamb = False
    nlamb = len(lamb)
   
  Coef, DF, Btiter, Btenter, ITER, endmodelno, Lamb, Stops, openmp = pga(x,
                                                            Z,
                                                            penalty,
                                                            zeta,
                                                            c,
                                                            lamb, nlamb, makelamb, lamb_min_ratio,
                                                            penalty_factor,
                                                            reltol,
                                                            maxiter,
                                                            steps,
                                                            btmax,
                                                            M,
                                                            tau,
                                                            nu,
                                                            alg,
                                                            fix_array,
                                                            lse,
                                                            Lmin,
                                                            nthreads          ,
                                                            wave,
                                                            J,
                                                            dimglam,
                                                            wf)


  if(np.mean(Stops[1, ]) > 0):
    zs = np.where(Stops[1, ] != 0)
    print("Warning: Maximum number of inner iterations (" + str(maxiter) + ") reached for model no." + str(endmodelno[zs] + 1) + " for zeta(s)" + str(zeta[zs]))

  if(np.mean(Stops[2, ]) > 0):
    zs = np.where(Stops[2, ] != 0)
    print("Warning: Maximum number of backtraking steps reached for model no." + str(endmodelno[zs] + 1) + " for zeta(s)" + str(zeta[zs]))

  if(openmp == 1):
    print("Note: Multithreading was used with " + str(nthreads) + " threads")
 
  
  maxiterpossible = np.sum(np.where(ITER > 0, 1, 0))
  maxiterreached = np.sum(np.where(ITER >= (maxiter - 1), 1, 0))
  
  if(maxiterreached > 0):
     print("Warning: Maximum number of inner iterations (" + str(maxiter) + ") reached " + str(maxiterreached) + " time(s) out of " + str(maxiterpossible) + " possible")
   
  if(num_of_zeta > 1):
    #Obj = []
    iter = [None] *  num_of_zeta
    coef = [None] *  num_of_zeta
    lamb_out = [None] *  num_of_zeta
    df = [None] *  num_of_zeta
    for z in range(num_of_zeta):
      coef[z] = Coef[: , 0:int(endmodelno[z]) + 1, z] / scale_y
      lamb_out[z] = Lamb[0:int(endmodelno[z]) + 1, z] / scale_y
      df[z] = DF[0:int(endmodelno[z]) + 1, z]
      iter[z] = ITER[0:int(endmodelno[z]) + 1, z]
     ## Obj[[z]] = res$Obj[, 1:int(endmodelno[z]) ,1]
  else:
    coef = Coef[: , 0:int(endmodelno) + 1, 0] / scale_y
    lamb_out = Lamb[0:int(endmodelno) + 1, 0] / scale_y
    df = DF[0:int(endmodelno) + 1, 0]
    iter = ITER[0:int(endmodelno) + 1, 0]
   ## Obj = OBJ[:, 0:int(endmodelno) ,0]

  if(fix_array):
    if(wave == 0):
      spec = str(penalty) + "-penalized smme model with " + str(G) + " groups and fixed " + str(dimglam) + "D-tensor " + "custom "  + "design"
    else:
      spec = str(penalty) + "-penalized smme model with " + str(G) + " groups and fixed " + str(dimglam) + "D-tensor " + wf + "-wavelet "  + "design"
  else:
    spec = str(penalty) + "-penalized smme model with varying inputs across the " + str(G) + " groups" 
   
  out = _smme_dict({"df":df, "lamb":lamb_out, "zeta":zeta, "spec":spec, "iter":iter})
  out["wf"] = wf
  out["spec"] = spec  
  out["zeta"] = zeta
  out["coef"] = coef
  out["lamb"] = lamb_out
  out["df"] = df
  out["diagnostics"] = {"iter" : iter, "bt_iter" : Btiter, "bt_enter" : Btenter}
  #out["Obj"] = Obj

  if(fix_array == 1):
    out["dimcoef"] = [p1, p2, p3]#[0:dimglam]
    out["dimobs"] = [n1, n2, n3]#[0:dimglam]
    out["dimmodel"] = dimglam
  else:
    out["dimcoef"] = p
    out["dimobs"] = n
    out["dimmodel"] = None

  out["endmod"] = endmodelno

  return out

def predict(fit, x):

 r""" Make predictions from a fitted smme model.
 
   Parameters
   ----------
   fit : smme_dict
       The output from a ``pysmme.tools.softmaximin`` call
 
   x : list, np.array or string
       An object that should be like the input to the ``pysmme.tools.softmaximin`` call that 
       produced the object ``fit``. For general  models a matrix
       with column dimension equal to that of  the original input. 
       For array models with custom design a list like the one supplied to ``softmaximin`` to produce ``fit``
       and for a wavelet design the name of the wavelet used to produce ``fit``. 
 
   Returns
   -------
   list     
        A list of length ``len(zeta)``. If ``x`` is a :math:`k \times p` matrix 
        each list item is a :math:`k \times m_\zeta` matrix containing the linear
        predictors computed for each ``lamb``. If ``x`` is a string or  a
        list of matrices and ``fit["dimmodel"] = d``,  each list item is a :math:`d + 1` array 
        containing predictions computed for each ``lamb``.
 
   Notes
   -----
    Given input ``fit`` and ``x``, this function computes the linear predictors
    using the fitted model coefficients supplied in  ``fit``  produced by  
    ``softmaximin``. If ``fit`` is the result of fitting general type model  
    ``x`` should be a :math:`k \times p` matrix (:math:`p` is the number of model
    coefficients and :math:`k` is the number of new data points). 
    If ``fit`` is the result of fitting a model with tensor design, ``x`` should 
    be a list containing :math:`k_i \times p_i, i = 1, 2, 3` matrices 
    (:math:`k_i` is the number of new marginal 
    data points in the :math:`i` th dimension) or a string indicating the wavelet 
    used to produce ``fit``.
   
   Examples
   --------
   
  #array data 
  ##size of example

  >>> G = 3; 
  >>> n = np.array([65, 26, 13])
  >>> p = np.array([13, 5, 4])
  
  ##marginal design matrices (Kronecker components)

  >>> x = [None] * 3 
  >>> for i in range(len(x)):
  >>> x[i] = np.random.normal(0, 1, (n[i], p[i]))
       
  ##common features and effects

  >>> common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
  >>> common_effects = np.random.normal(size = np.prod(p)) * common_features
  
  ##group response

  >>> y = np.zeros((n[0], n[1], n[2], G))
  >>> for g in range(G):
  >>>     bg = np.random.normal(0, 0.1, np.prod(p)) * (1 - common_features) + common_effects
  >>>     mu = RH(x[2], RH(x[1], RH(x[0], np.reshape(bg, (p[0], p[1], p[2]), "F") )))
  >>>     y[:, :, :, g] = np.random.normal(0, 1, (n)) + mu
  
  ##fit model for range of lambda and zeta
  
  >>> zeta = np.array([0.1, 1, 10, 100])
  >>> fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
  >>> yhat = predict(fit, x)
   
  #Array data and wavelets
  ##size of example

  >>> G = 5; 
  >>> p = n = np.array([2**2, 2**3, 2**4])
  
  ##wavelet design

  >>> x = "la8"

  ##common features and effects
  
  >>> common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
  >>> common_effects = np.random.normal(size = np.prod(p)) * common_features
  
  ##group response

  >>> y = np.zeros((n[0], n[1], n[2], G))
  >>> for g in range(G):
  >>>     bg = np.random.normal(0, 0.1, np.prod(p)) * (1 - common_features) + common_effects
  >>>     mu = iwt(np.reshape(bg, (p[0], p[1], p[2]), "F"))
  >>>     y[:, :, :, g] = np.random.normal(0, 1, (n)) + mu
  
  ##fit model for range of lambda and zeta
  
  >>> zeta = np.array([0.1, 1, 10, 100])
  >>> fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
  >>> modelno = 10
  >>> zetano = 2
  >>> yhat = predict(fit, x)
  >>> yhat[zetano][:,:,:, modelno]
  
  #Non-array data
  ##size of example

  >>> G = 10
  >>> n = np.random.choice(np.arange(100,500,1), G) #sample(100:500, G); 
  >>> p = 60
  >>> x = [None] * G
  
  ##group design matrices

  >>> for i in range(len(x)):
  >>> x[i] = np.random.normal(0, 1, (n[i], p))
  
  ##common features and effects

  >>> common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
  >>> common_effects = np.random.normal(size = np.prod(p)) * common_features
  
  ##group response

  >>> y = [None] * G
  >>> for g in range(G):
  >>>     bg = np.random.normal(0, 0.5, np.prod(p)) * (1 - common_features) + common_effects
  >>>     mu = np.matmul(x[g], bg)
  >>>     y[g] = np.random.normal(0, 1, n[g]) + mu
  
  ##fit model for range of lamb and zeta
  
  >>> fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
  >>> yhat = predict(fit, x)

   """    
  
 n = fit["dimobs"]
 p = fit["dimcoef"]
 dim_model = fit["dimmodel"] #nNone f0r general
 nzeta = len(fit["zeta"])
 out = [None] * nzeta

 if(type(x) == list or type(x) == str): ##array model
   if(not len(x) == dim_model):
    return "the length of x must be equal to the dimension of the model!"
   if(type(x) == list): ##custom non wavelet model 
    nx = [None] * len(x)
    px = [None] * len(x)
    for i in range(len(x)):
      nx[i] = x[i].shape[0]
      px[i] = x[i].shape[1]
    if(not px == p):
      return "the parameter dimension of the supplied data is not equal to the parameter dimension of the model "
    if(dim_model == 1):
     x.append(np.ones((1, 1)))
     x.append(np.ones((1, 1)))
    elif(dim_model == 2):
     x.append(np.ones((1, 1)))
   else: #wavelet
     nx = n
     px = p  

   for z in range(nzeta):  
     nlambda = len(fit["lamb"][z])
     res = np.zeros((nx[0], nx[1], nx[2], nlambda), order = "F")
     for m in range(nlambda):
      beta = np.reshape(fit["coef"][z][:, m], (px[0], px[1], px[2]), order = "F")
      if(type(x) == str):
       res[:,:,:, m] = iwt(beta, wf = x)
      else:
       res[:, :, :, m] = RH(x[2], RH(x[1], RH(x[0], beta)))
     out[z] = res
 elif(type(x) == np.array): #varying design general model
   nx = x.shape[0]
   px = x.shape[1]
   if(not px == p):
      return "column dimension of the new data x is not equal to the number of coefficients in the model"
   
   for z in range(nzeta):
    nlambda = len(fit["lamb"][z])
    res = np.zeros((nx, nlambda), order = "F")
    for m in range(nlambda):
     res[:, m] = np.maltmul(x, fit["coef"][: , m])...?z
    out[z] = res
   
  #return("dimension of new data inconsistent with existing data"))
 
 return out



"""
This module contains various transforms for computing fast matrix vector products in specific situations.

"""

import numpy as np
import pysmme._smme

def wt(x, wf = "la8", J = None):
 r"""Discrete wavelet transform.

   This function performs a level J wavelet transform of the input array (1d, 2d, or 3d) 
   using the pyramid algorithm (Mallat 1989). Implemented in C by Brandon Whithcer.

   Parameters
   ----------
   x : np.array
       A 1, 2, or 3 dimensional data array. The size of each dimension must be dyadic.

   wf : string
       The type of wavelet family used. Options are 
       ``"haar", "d4", "??", "mb4", "fk4", "d6", "fk6", "d8", "fk8", "la8", "mb8", "bl14",
       "fk14", "d16", "la16", "mb16", "la20", "bl20", "fk22", "mb24"``

   J : int
       J is the level (depth) of the decomposition. For default None the max
       depth is used and  ``wt(x)`` is equal to multiplying ``x`` with the
       corresponding wavelet matrix.

   Returns
   -------
   np.array
      Array with shape identical to input ``x`` containing the transform values. 

   Notes
   -----
   This is a C++/R wrapper function for a C implementation of the
   discrete wavelet transform by Brandon Whitcher licensed under the BSD 3 license
   https://cran.r-project.org/web/licenses/BSD_3_clause, see the Waveslim package;
   Percival and Walden (2000); Gencay, Selcuk and Whitcher (2001).
   Given a data array (1d, 2d or 3d) with dyadic dimensions sizes this transform 
   is computed efficiently via the pyramid algorithm.

   This functionality is used in the computations underlying ``pysmme.tools.softmaximin``
   to perform multiplications involving the wavelet (design) matrix efficiently.

   References
   ----------
   .. [3]  Gencay, R., F. Selcuk and B. Whitcher (2001) An Introduction to Wavelets and
      Other Filtering Methods in Finance and Economics, Academic Press.

   .. [4] Mallat, S. G. (1989) A theory for multiresolution signal decomposition: the
      wavelet representation, IEEE Transactions on Pattern Analysis and Machine
      Intelligence, 11, No. 7, 674-693.

   .. [5] Percival, D. B. and A. T. Walden (2000) Wavelet Methods for Time Series
       Analysis, Cambridge University Press.

   Examples
   --------
   >>> d = np.reshape(np.arange(1, 2**3 + 1,1), (2, 2, 2), order = "F")
   >>> d1 = wt(d)
   >>> d2 = np.array([[[ 1.41421356e+00,  4.16333634e-17],
        [ 5.65685425e+00, -3.33644647e-16]],
       [[ 2.82842712e+00, -2.77555756e-17],
        [-2.64953102e-16,  1.27279221e+01]]])
   >>> iwt(d2)     
   """
 d = np.array(x.shape)
 dim = sum(d > 1)
 p1 = d[0]
 if(J == None):
  J = int(np.log2(min(d[d > 1])))
 
 p3 = 1 
 p2 = 1
 if(dim > 1):
  p2 = d[1]
 if(dim > 2):
  p3 = d[2]
 X = np.reshape(x, (p1, p2 * p3), order = "F")

 out = pysmme._smme.WT(X, dim, wf, J, p1, p2, p3) #p1 x p2 * p3

 return np.reshape(out, (p1, p2, p3), order = "F")

def iwt(x, wf = "la8", J = None):
 r"""Discrete inverse wavelet transform.

  This function performs a level J wavelet transform of a dyadic input array (1d, 2d, or 3d) 
  using the pyramid algorithm (Mallat 1989). Implemented in C by Brandon Whithcer.

  Parameters
  ----------
  x : np.array
      A 1, 2, or 3 dimensional data array. The size of each dimension must be dyadic.

  wf : string
      The type of wavelet family used. Options are ``"haar", "d4", "??","mb4", "fk4", "d6", "fk6", 
      "d8","fk8", "la8","mb8","bl14","fk14", "d16","la16","mb16", "la20","bl20","fk22", "mb24"``

  J : int
      The level (depth) of the decomposition. For default ``None`` the max
      depth is used and  ``wt(x)`` is equal to multiplying ``x`` with the
      corresponding inverse wavelet transform matrix.

  Returns
  -------
  np.array
      Array with shape identical to input ``x`` containing the transform values. 

  Notes
  -----
  This is a C++/R wrapper function for a C implementation of the inverse
  discrete wavelet transform by Brandon Whitcher licensed under the BSD 3 license
  https://cran.r-project.org/web/licenses/BSD_3_clause, see the Waveslim package;
  Percival and Walden (2000); Gencay, Selcuk and Whitcher (2001).
  Given a data array (1d, 2d or 3d) with dyadic dimensions sizes this transform 
  is computed efficiently via the pyramid algorithm.

  This functionality is used in the computations underlying ``softmaximin``
  to perform multiplications involving the wavelet (design) matrix efficiently.

  References
  ----------

  Examples
  --------

  """    
 d = np.array(x.shape)
 dim = sum(d > 1)
 p1 = d[0]
 if(J == None):
  J = int(np.log2(min(d[d > 1])))
 p3 = 1 
 p2 = 1
 if(dim > 1):
  p2 = d[1]
 if(dim > 2):
  p3 = d[2]
 X = np.reshape(x, (p1, p2 * p3), order = "F")
 out = pysmme._smme.IWT(X, dim, wf, J, p1, p2, p3) #p1 x p2 * p3
 return np.reshape(out, (p1, p2, p3), order = "F")

def H(M, A):
  d = A.shape
  Amat = np.reshape(A, [d[0], np.prod(d[1:len(d)])], "F")
  MAmat = np.matmul(M, Amat)
  newdim = list(d[1:len(d)])
  newdim.insert(0, MAmat.shape[0])
  return np.reshape(MAmat, newdim, "F")
  
# Rotation of an array A
def Rotate(A):
  d = np.arange(1, len(A.shape) + 1, 1)
  d[len(A.shape) - 1] = 0
  return np.transpose(A, d)

def RH(M, A):
  r"""The Rotated H-transform of a 3d Array by a Matrix.

  This function is an implementation of the :math:`\rho`-operator found in
  (Currie et al, 2006). It forms the basis of the GLAM arithmetic.

  Parameters
  ----------
  M : np.array
      A :math:`n \times p_1` matrix.

  A : np.array
      A 3d array of size :math:`p_1 \times p_2 \times p_3`.

  Returns
  -------
  np.array
      A 3d array of size :math:`p_2 \times p_3 \times n`.

  Notes
  -----
  For details see (Currie et al, 2006) [2]_. Note that this particular implementation
  is not used in the  routines underlying the optimization procedure.

  References
  ----------
  .. [2]  Currie, I. D., M. Durban, and P. H. C. Eilers (2006). Generalized linear
     array models with applications to multidimensional smoothing.
     Journal of the Royal Statistical Society. Series B. 68, 
     259-280. url = http://dx.doi.org/10.1111/j.1467-9868.2006.00543.x.

  Examples
  --------
  
  >>> n1 = 15; n2 = 4; n3 = 3; p1 = 12; p2 = 3; p3 = 4
  >>> ###marginal design matrices (Kronecker components)
  >>> X1 = np.random.normal(0, 1, (n1, p1))
  >>> X2 = np.random.normal(0, 1, (n2, p2))
  >>> X3 = np.random.normal(0, 1, (n3, p3))
  >>> A = np.random.normal(0, 1, (p1, p2, p3))
  >>> R1 = RH(X3, RH(X2, RH(X1, A)))
  >>> R2 = np.matmul(np.kron(X3, np.kron(X2, X1)), np.reshape(A, [p1 * p2 * p3, 1], "F"))
  >>> max(abs(np.reshape(R1, [n1 * n2 * n3, 1], "F") - R2)) 
  """    
  return Rotate(H(M, A))


import matplotlib.pyplot as plt
import numpy as np
from pysmme.tools import softmaximin
from pysmme.tools import predict
from pysmme.transforms import RH, iwt, wt
#array data
##size of example
G = 3; 
n = np.array([65, 26, 13])
p = np.array([13, 5, 4])

##marginal design matrices (Kronecker components)
x = [None] * 3 
for i in range(len(x)):
    x[i] = np.random.normal(0, 1, (n[i], p[i]))
     
##common features and effects
common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
common_effects = np.random.normal(size = np.prod(p)) * common_features

##group response
y = np.zeros((n[0], n[1], n[2], G))
for g in range(G):
    bg = np.random.normal(0, 0.1, np.prod(p)) * (1 - common_features) + common_effects
    mu = RH(x[2], RH(x[1], RH(x[0], np.reshape(bg, (p[0], p[1], p[2]), "F") )))
    y[:, :, :, g] = np.random.normal(0, 1, (n)) + mu

##fit model for range of lambda and zeta
zeta = np.array([0.1, 1, 10, 100])
#zeta = np.array(10)

fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
modelno = 10
zetano = 2
betahat = fit["coef"][zetano][:, modelno]
 
f, ax = plt.subplots(1)
ax.plot(common_effects, "r+")
ax.plot(betahat)
plt.show()

#Array data and wavelets
##size of example
G = 5; 
p = n = np.array([2**2, 2**3, 2**3])
#wavelet design
x = "la8"

##common features and effects
common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
common_effects = np.random.normal(size = np.prod(p)) * common_features

##group response
y = np.zeros((n[0], n[1], n[2], G))
for g in range(G):
    bg = np.random.normal(0, 0.1, np.prod(p)) * (1 - common_features) + common_effects
    mu = iwt(np.reshape(bg, (p[0], p[1], p[2]), order = "F"))
    y[:, :, :, g] = np.random.normal(0, 1, (n)) + mu

##fit model for range of lambda and zeta
zeta = np.array([0.1, 1, 10, 100])
fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
modelno = 5
zetano = 3
betahat = fit["coef"][zetano][:, modelno]
 
f, ax = plt.subplots(1)
ax.plot(common_effects, "r+")
ax.plot(betahat)
plt.show() 

yhat = predict(fit, "la8")
plt.plot(yhat[zetano][:,2,2,modelno])

##Non-array data
##size of example
G = 10
n = np.random.choice(np.arange(100,500,1), G) #sample(100:500, G); 
p = 60
x = [None] * G

 ##group design matrices
 #for(g in 1:G){x[[g]] = matrix(rnorm(n[g] * p), n[g], p)}
for i in range(len(x)):
    x[i] = np.random.normal(0, 1, (n[i], p))

 ##common features and effects
common_features = np.random.binomial(1, 0.1, np.prod(p)) #sparsity of common effects
common_effects = np.random.normal(size = np.prod(p)) * common_features

##group response
y = [None] * G
for g in range(G):
    bg = np.random.normal(0, 0.5, np.prod(p)) * (1 - common_features) + common_effects
    mu = np.matmul(x[g], bg)
    y[g] = np.random.normal(0, 1, n[g]) + mu

 ##fit model for range of lamb and zeta
fit = softmaximin(y, x, zeta = zeta, penalty = "lasso", alg = "npg")
betahat = fit["coef"]

 ##estimated common effects for specific lamb and zeta
modelno = 6 
zetano = 2
f, ax = plt.subplots(1)
ax.plot(common_effects, "r+")
ax.plot(betahat[zetano][:, modelno])
plt.show() 

 
 

""" d=np.reshape(np.arange(1,65,1), (4, 4, 4), order = "F")

d=np.reshape(np.array([1.042495e-02,  2.818002e+00  ,1.042495e-02 , 2.818002e+00  ,1.042495e-02
,   2.818002e+00,  1.042495e-02,  2.818002e+00 , 4.169982e-02,  4.169982e-02
, 1.127201e+01  ,1.127201e+01 , 4.169982e-02  ,4.169982e-02 , 1.127201e+01
,  1.127201e+01 , 1.667993e-01,  1.667993e-01 , 1.667993e-01,  1.667993e-01
,  4.508803e+01 , 4.508803e+01,  4.508803e+01 , 4.508803e+01, -5.645764e-15
, -2.847085e-16 ,-2.921765e-15, -2.356061e-15 ,-1.146342e-15,  3.043347e-16
, -1.611225e-15 ,-5.780490e-16,  6.926968e-16 , 4.996004e-16, -1.800535e-15
,-4.440892e-16 ,-4.265902e-15 ,-8.604228e-16 ,-3.066666e-15 ,-6.383782e-16
,  2.360959e-15, -6.739401e-16, -8.770762e-15, -9.325873e-15 , 9.256484e-15
,  1.354515e-14,  4.440892e-15, -1.199041e-14,  9.578527e-16 , 4.318396e-16
, -4.272894e-16,  8.997692e-16, -3.608994e-15,  1.176928e-16, -2.251106e-15
, -3.993022e-16, -6.945144e+00, -2.778057e+01, -1.111223e+02,  1.703550e-14
,  8.326673e-15, -2.398082e-14, -1.711401e-14 , 2.600000e+02]),(4,4,4), order = "F")

d1=iwt(d)
wt(d1)

 """


import os
import re
import subprocess
import sys

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection & inclusion of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        # In this example, we pass in the version to C++. You might not need to.
        cmake_args += [f"-DEXAMPLE_VERSION_INFO={self.distribution.get_version()}"]

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator:
                try:
                    import ninja  # noqa: F401

                    cmake_args += ["-GNinja"]
                except ImportError:
                    pass

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}"
                ]
                build_args += ["--config", cfg]

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += [f"-j{self.parallel}"]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )

classifiers = ["Development Status :: 5 - Production/Stable",
"Intended Audience :: Education",
"Operating System :: MacOS",
"License :: OSI Approved :: MIT License",
"Programming Language :: Python :: 3",]

long_des = """Efficient C++ procedure for solving the Lasso or SCAD penalized soft maximin problem. 

This is a C++ implementation of two proximal
gradient based algorithms (NPG and FISTA) that solve different forms of the soft
maximin problem from Lund et al., 2022 see https://doi.org/10.1111/sjos.12580. 
1) For general group specific design the soft maximin problem is solved using 
the NPG algorithm.
2) For fixed identical design across groups, the  soft maximin problem is solved using  
either the FISTA algorithm or the NPG algorithm in the following two cases:
i) For a tensor structured design matrix the algorithms use array arithmetic  to 
avoid the design matrix and speed computations ii) For a wavelet based design 
matrix the algorithms use the pyramid algorithm to avoid the design matrix and speed up
computations. Multi-threading is possible when openMP is available."""

# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.
setup(
    name="pysmme",
    version="1.0",
    license= "mit",
    classifiers=classifiers,
    author="Adam Lund",
    url="https://adam-lund.github.io",
    author_email="adam.lund@math.ku.dk",
    description="Soft maximin estimation in Python",
    long_description= long_des,
    ext_modules=[CMakeExtension("pysmme._smme")],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
   # extras_require={"test": ["pytest>=6.0"]},
    python_requires=">=3.6",
        packages=['pysmme']
)
